# $Id$
#
#  Copyright (C) 2010 Gianluca Sforna
#
#  All Rights Reserved
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.

import math

class CanvasBase:
    """Base class for specialized canvas backends"""

    def addCanvasLine(self, p1, p2, color=(0, 0, 0), color2=None, **kwargs):
        """Draw a single line on the canvas

            This function will draw a line between `p1` and `p2` with the
            given `color`.
            If `color2` is specified, it will be used to draw the second half
            of the segment
        """
        raise NotImplementedError('This should be implemented')

    def addCanvasText(self, text, pos, font, color=(0, 0, 0), **kwargs):
        """Draw some text

           The provided `text` is drawn at position `pos` using the given
           `font` and the chosen `color`.
        """
        raise NotImplementedError('This should be implemented')

    def addCanvasPolygon(self, ps, color=(0, 0 ,0), **kwargs):
        """Draw a polygon

           Draw a polygon identified by vertexes given in `ps` using
           the given `color` 
        """
        raise NotImplementedError('This should be implemented')

    def addCanvasDashedWedge(self, p1, p2, p3, dash=(2, 2), 
                             color=(0, 0, 0), color2=None, **kwargs):
        """Draw a dashed wedge

           The wedge is identified by the three points `p1`, `p2`, and `p3`. 
           It will be drawn using the given `color`; if `color2` is specified
           it will be used for the second half of the wedge 

           TODO: fix comment, I'm not sure what `dash` does

        """
        raise NotImplementedError('This should be implemented')

    def flush(self):
        """Complete any remaining draw operation

           This is supposed to be the last operation on the canvas before
           saving it
        """
        raise NotImplementedError('This should be implemented')

    def _getLinePoints(self, p1, p2, dash):
        x1, y1 = p1
        x2, y2 = p2
        dx = x2 - x1
        dy = y2 - y1
        lineLen = math.sqrt(dx * dx + dy * dy)
        theta = math.atan2(dy, dx)
        cosT = math.cos(theta)
        sinT = math.sin(theta)

        pos = (x1, y1)
        pts = [pos]
        dist = 0
        currDash = 0
        while dist < lineLen:
            currL = dash[currDash % len(dash)]
            if (dist + currL > lineLen): currL = lineLen - dist
            endP = (pos[0] + currL * cosT, pos[1] + currL * sinT)
            pts.append(endP)
            pos = endP
            dist += currL
            currDash += 1
        return pts



# -*- coding: utf-8 -*-
# Author Karol Sikora <karol.sikora@laboratorium.ee>, (c) 2012
# Author Michal Nowotka <mmmnow@gmail.com>, (c) 2013-2014

try:
    import cPickle as pickle
except ImportError:
    import pickle
import base64
import pymongo
from datetime import datetime, timedelta
from chembl_beaker.beaker.cache.backends.base import BaseCache
import zlib
import logging
import json
from chembl_beaker.beaker import config

#-----------------------------------------------------------------------------------------------------------------------

MAX_SIZE = 16000000

#-----------------------------------------------------------------------------------------------------------------------

class MongoDBCache(BaseCache):

    def __init__(self):
        super(MongoDBCache, self).__init__()
        self._host = config.get('mongo_host', 'localhost')
        self._port = int(config.get('mongo_port', 27017))
        self._database = config.get('mongo_db', 'beaker_cache')
        self._rshosts = config.get('mongo_rshosts')
        self._rsname = config.get('mongo_rsname')
        self._user = config.get('mongo_user', None)
        self._password = config.get('mongo_pass', None)
        self._socket_timeout_ms = config.get('mongo_socket_simeout_ms', None)
        self._connect_timeout_ms = config.get('mongo_connect_timeout_ms', 300)
        self.compression_level = config.get('mongo_compression_level', 6)
        self._tag_sets = json.loads(config.get('mongo_tag_sets', []))
        self._read_preference = config.get("mongo_read_preference")
        self._collection = config.get('mongo_collection', 'cache')
        self.log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------------------------------------------------

    def add(self, key, value):
        key = self.make_key(key)
        self._base_set('add', key, value)

#-----------------------------------------------------------------------------------------------------------------------

    def set(self, key, value):
        key = self.make_key(key)
        self._base_set('set', key, value)

#-----------------------------------------------------------------------------------------------------------------------

    def _base_set(self, mode, key, value, timeout=None):
        if not timeout:
            timeout = self.default_timeout
        now = datetime.utcnow()
        expires = now + timedelta(seconds=timeout)
        coll = self._get_collection()
        encoded = self._encode(value)
        document_size = len(encoded)
        count = coll.count()
        if count > self._max_entries:
            self._cull()
        data = coll.find_one({'_id': key})
        if data and (mode == 'set' or
                (mode == 'add' and data['expires'] > now)):
            raw = data.get('data')
            if raw and raw == encoded:
                coll.update({'_id': data['_id']}, {'$set': {'expires': expires}}, safe=True)
                return
            else:
                self._delete([key] + data.get('chunks', []))
        if document_size <= MAX_SIZE:
            coll.insert({'_id': key, 'data': encoded, 'expires': expires}, safe=True)
        else:
            chunks = []
            for i in xrange(0, document_size, MAX_SIZE):
                chunk = encoded[i:i+MAX_SIZE]
                aux_key = self.make_key(chunk)
                coll.insert({'_id': aux_key, 'data': chunk}, safe=True)
                chunks.append(aux_key)
            coll.insert({'_id': key, 'chunks': chunks, 'expires': expires}, safe=True)

#-----------------------------------------------------------------------------------------------------------------------

    def _decode(self, data):
        return pickle.loads(zlib.decompress(base64.decodestring(data)))

#-----------------------------------------------------------------------------------------------------------------------

    def _encode(self, data):
        return base64.encodestring(zlib.compress(pickle.dumps(data, pickle.HIGHEST_PROTOCOL), self.compression_level))

#-----------------------------------------------------------------------------------------------------------------------

    def get(self, key, default=None):
        coll = self._get_collection()
        key = self.make_key(key)
        now = datetime.utcnow()
        data = coll.find_one({'_id': key})
        if not data:
            return default
        if data['expires'] < now:
            coll.remove(data['_id'])
            return default
        raw = data.get('data')
        if not raw:
            chunks = data.get('chunks')
            if chunks:
                raw = ''
                for chunk in chunks:
                    raw += coll.find_one({'_id': chunk})['data']
            else:
                return default
        return self._decode(raw)

#-----------------------------------------------------------------------------------------------------------------------

    def delete(self, key):
        key = self.make_key(key)
        coll = self._get_collection()
        data = coll.find_one({'_id': key})
        if data:
            self._delete([key] + data.get('chunks', []))

#-----------------------------------------------------------------------------------------------------------------------

    def _delete(self, ids_to_remove):
        coll = self._get_collection()
        coll.remove({'_id': {'$in':ids_to_remove}})

#-----------------------------------------------------------------------------------------------------------------------

    def clear(self):
        coll = self._get_collection()
        coll.remove({})

#-----------------------------------------------------------------------------------------------------------------------

    def _cull(self):
        if self._cull_frequency == 0:
            self.clear()
            return
        coll = self._get_collection()
        coll.remove({'expires': {'$lte': datetime.utcnow()}})
        #TODO: implement more agressive cull

#-----------------------------------------------------------------------------------------------------------------------

    def _get_collection(self):
        if not getattr(self, '_coll', None):
            self._initialize_collection()
        return self._coll

#-----------------------------------------------------------------------------------------------------------------------

    def _initialize_collection(self):
        try:
            from gevent import monkey
            monkey.patch_socket()
        except ImportError:
            pass

        if self._rsname:
            self.connection = pymongo.MongoReplicaSetClient(self._rshosts, replicaSet=self._rsname,
                read_preference=getattr(pymongo.ReadPreference,self._read_preference, 'PRIMARY'),
                socketTimeoutMS=self._socket_timeout_ms, connectTimeoutMS=self._connect_timeout_ms,
                                                                            tag_sets=self._tag_sets)
        else:
            self.connection = pymongo.Connection(self._host, self._port)

        self._db = self.connection[self._database]
        if self._user and self._password:
            self._db.authenticate(self._user, self._password)
        self._coll= self._db[self._collection]

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

from chembl_beaker.beaker import config
from chembl_beaker.beaker.utils import import_class
from chembl_beaker.beaker.cache.backends.base import BaseCache

cache = None

cache_class = config.get('cache_backend')

if cache_class:
    try:
        cache = import_class(cache_class)()
        if not isinstance(cache, BaseCache):
            print "Configured cache class (%s) is not a BaseCache instance, skipping caching." % cache_class
            cache = None
    except ImportError:
        print 'Error importing %s' % cache_class





__author__ = 'mnowotka'

from collections import OrderedDict
from bottle import response, static_file
from chembl_beaker import __version__ as version
import json

from chembl_beaker.beaker import app, config
from chembl_beaker.beaker import STATIC_ROOT
from chembl_beaker.beaker import PARAM_REGEX

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/docs')
def docs():
    return static_file('docs.html', root=STATIC_ROOT)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/spore', methods = ['OPTIONS', 'GET'])
def spore():
    """Print available functions."""

    ret = {
        "version": version,
        "expected_status": [200],
        "name": "ChEMBL Beaker API live documentation",
        "methods": OrderedDict()
    }

    post_methods = dict()
    get_methods = dict()

    for route in app.routes:
        name = route.name
        method = route.method.upper()
        uname = "%s_%s" % (method, name)
        if not name or method not in ('GET', 'POST'):
            continue
#        if uname in ret["methods"]:
#            #print "powtorka dla %s" % uname
#            params = map(lambda x: x[1:-1].upper(), PARAM_REGEX.findall(route.rule))
#            intersection = list(set(params) & set(ret["methods"][uname]["required_params"]))
#            #print "intersection = %s" % intersection
#            diff = list(set(params) ^ set(ret["methods"][uname]["required_params"]))
#            #print "diff = %s" % diff
#            #print "optional params = %s" % list(set(ret["methods"][uname].get("optional_params", []) + diff))
#            ret["methods"][uname]["required_params"] = intersection
#            ret["methods"][uname]["optional_params"] = list(set(ret["methods"][uname].get("optional_params", []) + diff))
#            continue
        method_info = dict()
        method_info["description"] = route.callback.__doc__
        method_info["method"] = method
        method_info["formats"] = ['text']
        method_info["path"] = PARAM_REGEX.sub(lambda x: ':' + x.group(0)[1:-1].upper(), route.rule)
        method_info["required_params"] = map(lambda x: x[1:-1].upper(), PARAM_REGEX.findall(route.rule))
        if method.upper() == 'POST':
            post_methods[uname] = method_info
        else:
            get_methods[uname] = method_info
    ret["methods"].update([(x, get_methods[x]) for x in sorted(get_methods.keys())])
    ret["methods"].update([(x, post_methods[x]) for x in sorted(post_methods.keys())])
    response.content_type = 'application/json'
    return json.dumps(ret)

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

from lxml.etree import _ElementTree, Element, tostring
from lxml import objectify
from lxml.objectify import ObjectifiedElement, StringElement
import json
from chembl_beaker import __version__ as version
from datetime import datetime, date
import rdkit
from rdkit import Chem
from rdkit.Chem.rdchem import GetPeriodicTable
from StringIO import StringIO
import os
import getpass

USERNAME = ''
try:
    USERNAME = getpass.getuser()
except:
    pass

MOL_MARVIN_SCALE = 1.8666666517333335

molBondTypes = {
"SINGLE" : 1,
"DOUBLE" : 2,
"TRIPLE" : 3,
"AROMATIC" : 4,
"UNSPECIFIED" : 8,
}

charges = {
    0:0,
    1:3,
    2:2,
    3:1,
    4:0,
    5:-1,
    6:-2,
    7:-3,
}

PERIODIC_TABLE = GetPeriodicTable()

#-----------------------------------------------------------------------------------------------------------------------

class MarvinJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if type(obj) == _ElementTree:
            return obj.getroot()
        if type(obj) == ObjectifiedElement:
            if obj.tag == 'cml':
                return  obj.find('MDocument')
            if obj.tag == 'MDocument':
                return [ el for el in obj.iterchildren(tag='MChemicalStruct') ]
            if obj.tag == 'MChemicalStruct':
                return obj.find('molecule')
            if obj.tag == 'molecule':
                return {obj.get("molID"): {"atoms" : obj.find("atomArray"), "bonds": obj.find("bondArray")}}
            if obj.tag == 'bondArray':
                return [ bond for bond in obj.iterchildren(tag='bond') ]
            if obj.tag == 'atomArray':
                return [ atom for atom in obj.iterchildren(tag='atom') ]
            if obj.tag == 'bond':
                return {"atomRefs" : obj.get('atomRefs2').split(),
                        "order": int(obj.get('order')),
                        "stereo": obj.find("bondStereo")}
            if obj.tag == 'bondStereo':
                print 'ELEM'
        if type(obj) == StringElement:
            if obj.tag == 'atomArray':
                ids = obj.get('atomID').split()
                num_atoms = len(ids)
                zeros = [0] * num_atoms

                if obj.get('x2', False):
                    return {"atomID": ids,
                            "x": [float(x) for x in obj.get('x2').split()],
                            'y': [float(y) for y in obj.get('y2').split()],
                            'z': zeros,
                            'dim': 2,
                            'elementType': obj.get('elementType').split()}

                else:
                    return {"atomID": ids,
                            "x": [float(x) for x in obj.get('x3').split()],
                            'y': [float(y) for y in obj.get('y3').split()],
                            'z': [float(y) for y in obj.get('z3').split()],
                            'dim':3,
                            'elementType': obj.get('elementType').split()}

            if obj.tag == 'bond':
                return {"atomRefs" : obj.get('atomRefs2').split(), "order": int(obj.get('order'))}

            if obj.tag == 'atom':
                if obj.get('x2', False):
                    return {"id" : obj.get('id'),
                            "elementType": obj.get('elementType'),
                            "x": float(obj.get('x2')),
                            "y": float(obj.get('y2')),
                            "z": 0.0,
                            'dim':2,
                            'formalCharge': int(obj.get('formalCharge', 0)),
                            'isotope':int(obj.get('isotope', 0))
                    }
                else:
                    return {"id" : obj.get('id'),
                            "elementType": obj.get('elementType'),
                            "x": float(obj.get('x3')),
                            "y": float(obj.get('y3')),
                            "z": float(obj.get('z3')),
                            'dim':3,
                            'formalCharge': int(obj.get('formalCharge', 0)),
                            'isotope':int(obj.get('isotope', 0))}

            if obj.tag == 'bondStereo':
                convention = obj.get('convention')
                if convention and convention.upper() == 'MDL':
                    return int(obj.get('conventionValue', 0))
                if not obj.text:
                    return 0
                t = obj.text.strip().upper()
                if t == 'W':
                    return 1
                elif t == 'H':
                    return 6
                else:
                    return 0

        return json.JSONEncoder.default(self, obj)

#-----------------------------------------------------------------------------------------------------------------------

def _dict2Mol(obj, scale = 1.0):
    buffer = StringIO()
    data = date.today().strftime("%m%d%y%H%M")

    atoms = obj['atoms']
    if type(atoms) == dict:
        num_atoms = len(atoms['atomID'])
    else:
        num_atoms = len(atoms)
    if type(atoms) == dict:
        dim = str(atoms.get('dim', 2)) + 'D'
    else:
        dim = str(atoms[0].get('dim',2)) + 'D'

    buffer.write('\n')
    buffer.write('{:2.2}{:^8.8}{:10.10}{:2.2}{:2.2}{:10.10}{:12.12}{:6.6}\n'.format(
                                                                    USERNAME, 'beaker', data, dim, '', '', '', ''))
    buffer.write('\n')

    zeros = [0] * num_atoms
    buffer.write('{:3}{:3}{:3}{:3}{:3}{:3}          {} V2000\n'.format(
                                                            num_atoms, len(obj['bonds']), 0, 0, 0, 0, 1))

    if type(atoms) == dict:
        for i, atom in enumerate(atoms['elementType']):
            buffer.write('{:10,.4f}{:10,.4f}{:10,.4f} {:<2}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}\n'.format(
                          atoms.get('x', zeros)[i] / scale,
                          atoms.get('y', zeros)[i] / scale,
                          atoms.get('z', zeros)[i] / scale,
                          atom,
                          0 if not atoms.get('isotope') else atoms.get('isotope')[i] -
                                                             int(PERIODIC_TABLE.GetAtomicWeight(str(atom))),
                          charges.keys()[charges.values().index(atoms.get('formalCharge', zeros)[i])],
                                                                                                0,0,0,0,0,0,0,0,0,0))
    else:
        for i, atom in enumerate(atoms):
            elementType = atom.get('elementType')
            buffer.write('{:10,.4f}{:10,.4f}{:10,.4f} {:<2}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}{:3}\n'.format(
                          atom.get('x', 0) / scale,
                          atom.get('y', 0) / scale,
                          atom.get('z', 0) / scale,
                          elementType,
                          0 if not atom.get('isotope') else atom.get('isotope') -
                                                            int(PERIODIC_TABLE.GetAtomicWeight(str(elementType))),
                          charges.keys()[charges.values().index(atom.get('formalCharge', 0))],
                                                                                                0,0,0,0,0,0,0,0,0,0 ))

    for bond in obj['bonds']:
        first_atom = None
        second_atom = None
        first_id, second_id = bond['atomRefs']
        stereo = bond.get('stereo', 0)
        if type(atoms) == dict:
            for i, atomID in enumerate(atoms['atomID']):
                if first_id == atomID:
                    first_atom = i + 1
                if second_id == atomID:
                    second_atom = i + 1
                if first_atom is not None and second_atom is not None:
                    break
        else:
            for i, atom in enumerate(atoms):
                atomID = atom.get('id')
                if first_id == atomID:
                    first_atom = i + 1
                if second_id == atomID:
                    second_atom = i + 1
                if first_atom is not None and second_atom is not None:
                    break
        buffer.write('{:3}{:3}{:3}{:3}{:3}{:3}{:3}\n'.format(first_atom, second_atom, bond['order'], stereo, 0, 0, 0))
    buffer.write('M  END\n')
    return buffer.getvalue()

#-----------------------------------------------------------------------------------------------------------------------

def _jsonToMol(obj, scale = 1.0):
    return '$$$$\n'.join([_dict2Mol(x.values()[0], scale) for x in obj])

#-----------------------------------------------------------------------------------------------------------------------

def _molToJson(mol, scale = 1.0):
    atoms = []
    conformer = mol.GetConformer()
    is3D = conformer.Is3D()
    for atom in mol.GetAtoms():
        atom_data = {}
        idx = atom.GetIdx()
        point = conformer.GetAtomPosition(idx)
        x = point.x * scale
        y = point.y * scale
        if is3D:
            z = point.y * scale
            atom_data['x3'] = str(x)
            atom_data['y3'] = str(y)
            atom_data['z3'] = str(z)
        else:
            atom_data['x2'] = str(x)
            atom_data['y2'] = str(y)
        atom_data['id'] = "a%s" % (idx + 1)
        atom_data["elementType"] = atom.GetSymbol()
        formalCharge = atom.GetFormalCharge()
        if formalCharge:
            atom_data["formalCharge"] = str(formalCharge)
        isotope = atom.GetIsotope()
        if isotope:
            atom_data["isotope"] = str(isotope)
        atoms.append(atom_data)
    bonds = []
    for bond in mol.GetBonds():
        atomRefs = ["a%s" % (bond.GetBeginAtomIdx() + 1), "a%s" % (bond.GetEndAtomIdx() + 1)]
        order = molBondTypes[str(bond.GetBondType())]
        stereo = bond.GetBondDir()
        bonds.append({"atomRefs": atomRefs, "order":order, "stereo":stereo})
    return {"atoms":atoms, "bonds":bonds}

#-----------------------------------------------------------------------------------------------------------------------

def _molsToJson(mols, scale = 1.0):
    ret = []
    for idx, mol in enumerate(mols):
        ret.append({"m%s" % (idx + 1): _molToJson(mol, scale)})
    return ret

#-----------------------------------------------------------------------------------------------------------------------

def _dictToEtree(data, name=None, depth=0):
    element = None
    if depth == 0:
        element = Element('cml')
        element.append(_dictToEtree(data, name, depth + 1))
    elif depth == 1:
        element = Element('MDocument')
        element.append(_dictToEtree(data, name, depth + 1))
    elif depth == 2:
        element = Element('MChemicalStruct')
        for mol in data:
            element.append(_dictToEtree(mol, name, depth + 1))
    elif depth == 3:
        molID = data.keys()[0]
        val = data.values()[0]
        element = Element('molecule', molID=molID)
        element.append(_dictToEtree(val['atoms'], 'atoms', depth + 1))
        element.append(_dictToEtree(val['bonds'], 'bonds', depth + 1))
    elif depth == 4:
        if name == 'atoms':
            element = Element('atomArray')
            for atom in data:
                element.append(_dictToEtree(atom, 'atom', depth + 1))
        elif name == 'bonds':
            element = Element('bondArray')
            for bond in data:
                element.append(_dictToEtree(bond, 'bond', depth + 1))
    elif depth == 5:
        if name == 'bond':
            kwargs = {}
            kwargs['atomRefs2'] = ' '.join(data['atomRefs'])
            kwargs['order'] = str(data['order'])
            element = Element('bond', **kwargs)
            stereo = data.get('stereo')
            if stereo:
                element.append(_dictToEtree(stereo, 'bondStereo', depth + 1))
        elif name == 'atom':
            element = Element('atom', **data)
    elif depth == 6:
        element = Element('bondStereo')
        if data == rdkit.Chem.rdchem.BondDir.BEGINWEDGE:
            element.text = 'W'
        elif data == rdkit.Chem.rdchem.BondDir.BEGINDASH:
            element.text = 'H'
        elif data == rdkit.Chem.rdchem.BondDir.UNKNOWN:
            element.attrib["convention"] = "MDL"
            element.attrib["conventionValue"] = "4"
    return element

#-----------------------------------------------------------------------------------------------------------------------

def _dataToXml(data, options=None):
    options = options or {}
    return tostring(_dictToEtree(data, options), xml_declaration=False, encoding='utf-8')

#-----------------------------------------------------------------------------------------------------------------------

def MolToMarvin(mol):
    if mol.endswith('.mol') and os.path.exists(mol):
        mol = Chem.MolFromMolFile(mol, False, False, False)
    else:
        mol = Chem.MolFromMolBlock(mol, False, False, False)
    js = _molsToJson([mol], MOL_MARVIN_SCALE)
    return _dataToXml(js)

#-----------------------------------------------------------------------------------------------------------------------

def SDFToMarvin(mol):
    mols = []
    if mol.endswith('.sdf') and os.path.exists(mol):
        suppl = Chem.SDMolSupplier('mol')
        mols = [mol for mol in suppl]

    js = _molsToJson(mols, MOL_MARVIN_SCALE)
    return _dataToXml(js)

#-----------------------------------------------------------------------------------------------------------------------

def MarvinToMol(marvin):
    if marvin.endswith('.mrv') and os.path.exists(marvin):
        f = open(marvin)
    else:
        f = StringIO(marvin)
    tree = objectify.parse(f)
    js = MarvinJSONEncoder().encode(tree)
    return _jsonToMol(json.loads(js), MOL_MARVIN_SCALE)

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

from rdkit import Chem
from rdkit.Chem import AllChem
from chembl_beaker.beaker.utils.functional import _apply
from chembl_beaker.beaker.utils.io import _parseMolData, _getSDFString, _parseSMILESData

#-----------------------------------------------------------------------------------------------------------------------

def _2D23D(mol, multi, mmff=False):
    molH = Chem.AddHs(mol)
    if mmff:
        optiFunc = AllChem.MMFFOptimizeMolecule
    else:
        optiFunc = AllChem.UFFOptimizeMolecule
    if multi:
        confIds=AllChem.EmbedMultipleConfs(molH, multi)
        for confId in confIds:
          optiFunc(molH, confId=confId)
    else:
        AllChem.EmbedMolecule(molH)
        optiFunc(molH)

    return molH

#-----------------------------------------------------------------------------------------------------------------------

def _ctab23D(data, multi, mmff):
    mols = _parseMolData(data)
    optimisedMols = _apply(mols, _2D23D, multi, mmff)
    return _getSDFString(optimisedMols)

#-----------------------------------------------------------------------------------------------------------------------

def _smiles23D(data, multi, mmff):
    mols = _parseSMILESData(data)
    optimisedMols = _apply(mols, _2D23D, multi, mmff)
    return _getSDFString(optimisedMols)

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

from rdkit import Chem
from rdkit.Chem import AllChem
from chembl_beaker.beaker.core_apps.marvin.MarvinJSONEncoder import MolToMarvin, MarvinToMol
from chembl_beaker.beaker.core_apps.D3Coords.impl import _2D23D

def _hydrogenize(block, hydro):
    mol = Chem.MolFromMolBlock(block)
    res = Chem.AddHs(mol, addCoords=True) if hydro else Chem.RemoveHs(mol)
    return MolToMarvin(Chem.MolToMolBlock(res))

#-----------------------------------------------------------------------------------------------------------------------

def _clean(mrv, dim=2):
    block = MarvinToMol(mrv)
    mol = Chem.MolFromMolBlock(block)
    if not mol:
        print "No mol for block:\n %s" % block
        return mrv
    AllChem.Compute2DCoords(mol, bondLength = 0.8)
    if dim == 3:
        mol = _2D23D(mol, True)
        mol = Chem.RemoveHs(mol)
    return MolToMarvin(Chem.MolToMolBlock(mol))

#-----------------------------------------------------------------------------------------------------------------------

def _stereoInfo(mrv):
    ret = {"headers":
               {"tetraHedral":
                    {"name":"tetraHedral","type":"COMPLEX","source":"CALCULATOR"},
                "doubleBond":
                    {"name":"doubleBond","type":"COMPLEX","source":"CALCULATOR"}
               },
           "tetraHedral":[],
           "doubleBond":[]
    }

    block = MarvinToMol(mrv)
    mol = Chem.MolFromMolBlock(block)
    if not mol:
        print "No mol for block:\n %s" % block
        return ret
    Chem.AssignStereochemistry(mol, flagPossibleStereoCenters=True, force=True)
    for atom in mol.GetAtoms():
        stereo = str(atom.GetChiralTag())
        atomIndex = atom.GetIdx()
        if str(atom.GetChiralTag()) != "CHI_UNSPECIFIED":
            if stereo == "CHI_TETRAHEDRAL_CW":
                chirality = "R"
            elif stereo == "CHI_TETRAHEDRAL_CCW":
                chirality = "S"
            else:
                chirality = "R/S"
            ret["tetraHedral"].append({"atomIndex":atomIndex,"chirality":chirality})
        elif atom.HasProp('_ChiralityPossible'):
            chirality = "R/S"
            ret["tetraHedral"].append({"atomIndex":atomIndex,"chirality":chirality})


    for bond in mol.GetBonds():
        stereo = str(bond.GetStereo())
        if stereo != "STEREONONE":
            atom1Index = bond.GetBeginAtomIdx()
            atom2Index = bond.GetEndAtomIdx()
            bondIndex = bond.GetIdx()
            if stereo == "STEREOANY":
                cistrans = "E/Z"
            elif stereo == "STEREOZ":
                cistrans = "Z"
            elif stereo == "STEREOE":
                cistrans = "E"
            ret["doubleBond"].append({"atom1Index": atom1Index,"atom2Index":atom2Index,"bondIndex":bondIndex,
                                                                                                "cistrans":cistrans})
    return ret

#-----------------------------------------------------------------------------------------------------------------------

def _molExport(structure, **kwargs):

    input_f = kwargs.get('input', None)
    output_f = kwargs.get('output', None)

    if not input_f:
        mol = _autoDetect(str(structure))
    elif input_f == 'mrv':
        mol = Chem.MolFromMolBlock(MarvinToMol(structure))
    elif input_f == 'smiles':
        mol = Chem.MolFromSmiles(str(structure))
    elif input_f == 'inchi':
        mol = Chem.MolFromInchi(str(structure))
    else:
        mol = Chem.MolFromMolBlock(structure)

    if not mol.GetNumConformers() or mol.GetConformer().Is3D():
        AllChem.Compute2DCoords(mol, bondLength = 0.8)

    if output_f == 'smiles':
        out_structure = Chem.MolToSmiles(mol)

    elif output_f == 'inchi':
        out_structure = Chem.MolToInchi(mol)

    elif output_f == 'inchikey':
        out_structure = Chem.InchiToInchiKey(Chem.MolToInchi(mol))

    elif output_f == 'mol':
        out_structure = Chem.MolToMolBlock(mol)

    elif output_f == 'sdf':
        out_structure = Chem.MolToMolBlock(mol) + '\n$$$$\n'

    elif output_f == 'mrv':
        out_structure = MolToMarvin(Chem.MolToMolBlock(mol))

    return {"structure": out_structure, "format":output_f, "contentUrl":"","contentBaseUrl":""}

#-----------------------------------------------------------------------------------------------------------------------

def _autoDetect(structure):

    if Chem.MolFromSmiles(structure):
        return Chem.MolFromSmiles(structure.strip())

    if Chem.MolFromMolBlock(structure):
        return Chem.MolFromMolBlock(structure)

    if Chem.INCHI_AVAILABLE and Chem.inchi.MolFromInchi(structure.strip(), True, True):
        return Chem.inchi.MolFromInchi(structure.strip(), True, True)

    if hasattr(Chem, 'MolFromSmarts') and Chem.MolFromSmarts(structure):
        return Chem.MolFromSmarts(structure)

    if hasattr(Chem, 'MolFromMol2Block') and Chem.MolFromMol2Block(structure):
        return Chem.MolFromMol2Block(structure)

    if hasattr(Chem, 'MolFromPDBBlock') and Chem.MolFromPDBBlock(structure):
        return Chem.MolFromPDBBlock(structure)

    if hasattr(Chem, 'MolFromTPLBlock') and Chem.MolFromTPLBlock(structure):
        return Chem.MolFromTPLBlock(structure)

    return None

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

from chembl_beaker.beaker import config

#-----------------------------------------------------------------------------------------------------------------------

class BaseThrottle(object):

    def __init__(self):
        self.hourly_rate_limit = config.get('throttle_hourly_rate_limit', 60 * 60)
        self.daily_rate_limit = config.get('throttle_daily_rate_limit', 60 * 60 * 24)
        self.key_hourly_rate_limit = config.get('throttle_key_hourly_rate_limit', 60 * 60 * 2)
        self.key_daily_rate_limit = config.get('throttle_key_daily_rate_limit', 60 * 60 * 24 * 2)

    def get_remaining_rates(self, identifier, type='IP'):
        raise NotImplementedError('subclasses of BaseThrottle must provide a get_remaining_rates method')

    def accessed(self, identifier, type='IP'):
        raise NotImplementedError('subclasses of BaseThrottle must provide an accessed method')

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

from chembl_beaker.beaker import app
from chembl_beaker.beaker.core_apps.similarityMaps.impl import _smiles2SimilarityMap, _sdf2SimilarityMap
from bottle import request, response
import base64

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/smiles2SimilarityMap/<smiles>', method=['OPTIONS', 'GET'], name="smiles2SimilarityMap")
def smiles2SimilarityMap(smiles):
    """
Generates similarity map, which is a way to visualize the atomic contributions to the similarity between molecules.
This method requires *exactly two SMILES*
    """

    data = base64.urlsafe_b64decode(smiles)
    response.content_type = 'image/png'
    ret = _smiles2SimilarityMap(data, request.params)
    if request.is_ajax:
        ret = base64.b64encode(ret)
    return ret

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/smiles2SimilarityMap', method=['OPTIONS', 'POST'], name="smiles2SimilarityMap")
def smiles2SimilarityMap():
    """
Generates similarity map, which is a way to visualize the atomic contributions to the similarity between molecules.
This method requires *exactly two SMILES*
    """

    data = request.body.read()
    response.content_type = 'image/png'
    ret = _smiles2SimilarityMap(data, request.params)
    if request.is_ajax:
        ret = base64.b64encode(ret)
    return ret

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/sdf2SimilarityMap/<ctab>', method=['OPTIONS', 'GET'], name="sdf2SimilarityMap")
def sdf2SimilarityMap(ctab):
    """
Generates similarity map, which is a way to visualize the atomic contributions to the similarity between molecules.
This method requires SDF containing *exactly two mols*
    """

    data = base64.urlsafe_b64decode(ctab)
    response.content_type = 'image/png'
    ret = _sdf2SimilarityMap(data, request.params)
    if request.is_ajax:
        ret = base64.b64encode(ret)
    return ret

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/sdf2SimilarityMap', method=['OPTIONS', 'POST'], name="sdf2SimilarityMap")
def sdf2SimilarityMap():
    """
Generates similarity map, which is a way to visualize the atomic contributions to the similarity between molecules.
This method requires SDF containing *exactly two mols*
    """

    data = request.body.read()
    response.content_type = 'image/png'
    ret = _sdf2SimilarityMap(data, request.params)
    if request.is_ajax:
        ret = base64.b64encode(ret)
    return ret

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

import bottle
from bottle import Bottle
import chembl_beaker
from chembl_beaker.beaker.utils import import_class
import re
import os
import sys
import json

#-----------------------------------------------------------------------------------------------------------------------

HTTP_CODES = bottle.HTTP_CODES.copy()
HTTP_CODES = dict((y,x) for x,y in HTTP_CODES.iteritems())
STATIC_ROOT = os.path.join(os.path.split(chembl_beaker.__file__)[0], 'static')
PARAM_REGEX = re.compile(r'<[^<>]+>')
DEFAULT_APPS = [
    "chembl_beaker.beaker",
    "chembl_beaker.beaker.core_apps.calculations",
    "chembl_beaker.beaker.core_apps.conversions",
    "chembl_beaker.beaker.core_apps.descriptors",
    "chembl_beaker.beaker.core_apps.fingerprints",
    "chembl_beaker.beaker.core_apps.marvin",
    "chembl_beaker.beaker.core_apps.mcs",
    "chembl_beaker.beaker.core_apps.osra",
    "chembl_beaker.beaker.core_apps.rasterImages",
    "chembl_beaker.beaker.core_apps.svgImages",
    "chembl_beaker.beaker.core_apps.jsonImages",
    "chembl_beaker.beaker.core_apps.standarisation",
    "chembl_beaker.beaker.core_apps.D3Coords",
    "chembl_beaker.beaker.core_apps.similarityMaps",
    "chembl_beaker.beaker.core_apps.autoDocs",
    ]

DEFAULT_PLUGINS = [
    'chembl_beaker.beaker.plugins.enableCors.EnableCors',
    'chembl_beaker.beaker.plugins.restrictions.Restrictions',
    'chembl_beaker.beaker.plugins.throttling.Throttling',
    'chembl_beaker.beaker.plugins.caching.Caching',
]

#-----------------------------------------------------------------------------------------------------------------------

def loadPlugins(app, plugins):
    if not plugins:
        plugins = DEFAULT_PLUGINS
    for plugin in plugins:
        try:
            plugin_class = import_class(plugin)
            app.install(plugin_class())
        except Exception as e:
            print "Failed to load plugin %s because of error %s" % (plugin, e.message)
            continue

#-----------------------------------------------------------------------------------------------------------------------

def loadApps(apps):
    if not apps:
        apps = DEFAULT_APPS
    for module in apps:
        try:
            __import__(module + ".views")
        except Exception as e:
            print "Loading module %s failed because of error: %s" % (module, e.message)
            continue

#-----------------------------------------------------------------------------------------------------------------------

app = Bottle()
config = app.config

#-----------------------------------------------------------------------------------------------------------------------

if not getattr(config, 'load_config'):

        py = sys.version_info
        py3k = py >= (3, 0, 0)

        if py3k:
            from configparser import ConfigParser
        else:
            from ConfigParser import SafeConfigParser as ConfigParser

        from bottle import ConfigDict

        def load_config(self, filename):
            ''' Load values from an *.ini style config file.

                If the config file contains sections, their names are used as
                namespaces for the values within. The two special sections
                ``DEFAULT`` and ``bottle`` refer to the root namespace (no prefix).
            '''
            conf = ConfigParser()
            conf.read(filename)
            for section in conf.sections():
                for key, value in conf.items(section):
                    if section not in ('DEFAULT', 'bottle'):
                        key = section + '.' + key
                    self[key] = value
            return self

        ConfigDict.load_config = load_config

#-----------------------------------------------------------------------------------------------------------------------





#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'mnowotka'

import sys

try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

setup(
    name='chembl_beaker',
    version='0.5.34',
    entry_points={
        'console_scripts': [
            'run_beaker=chembl_beaker.run_beaker:main']
    },
    author='Michal Nowotka',
    author_email='mnowotka@ebi.ac.uk',
    description='RDKit in the Bottle on Tornado',
    url='https://www.ebi.ac.uk/chembl/',
    license='CC BY-SA 3.0',
    packages=['chembl_beaker',
              'chembl_beaker.beaker',
              'chembl_beaker.beaker.plugins',
              'chembl_beaker.beaker.utils',
              'chembl_beaker.beaker.draw',
              'chembl_beaker.beaker.core_apps',
              'chembl_beaker.beaker.core_apps.autoDocs',
              'chembl_beaker.beaker.core_apps.calculations',
              'chembl_beaker.beaker.core_apps.conversions',
              'chembl_beaker.beaker.core_apps.descriptors',
              'chembl_beaker.beaker.core_apps.standarisation',
              'chembl_beaker.beaker.core_apps.D3Coords',
              'chembl_beaker.beaker.core_apps.fingerprints',
              'chembl_beaker.beaker.core_apps.jsonImages',
              'chembl_beaker.beaker.core_apps.marvin',
              'chembl_beaker.beaker.core_apps.mcs',
              'chembl_beaker.beaker.core_apps.osra',
              'chembl_beaker.beaker.core_apps.rasterImages',
              'chembl_beaker.beaker.core_apps.similarityMaps',
              'chembl_beaker.beaker.core_apps.svgImages',
              'chembl_beaker.beaker.cache',
              'chembl_beaker.beaker.cache.backends',
              'chembl_beaker.beaker.throttle',
              'chembl_beaker.beaker.throttle.backends',
              ],
    long_description=open('README.rst').read(),
    tests_require = ['Pillow', 'WebTest'],
    install_requires=['bottle>=0.11.6',
                      'tornado>=2.4',
                      'pytz',
                      'netaddr>=0.7.12'],
    package_data={
        'chembl_beaker': ['samples/*', 'static/*'],
        },
    include_package_data=False,
    classifiers=['Development Status :: 2 - Pre-Alpha',
                 'Environment :: Web Environment',
                 'Framework :: Bottle',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Scientific/Engineering :: Chemistry'],
    zip_safe=False,
)



"Memcached cache backend"

#-----------------------------------------------------------------------------------------------------------------------

import json
try:
    import cPickle as pickle
except ImportError:
    import pickle

from chembl_beaker.beaker.cache.backends.base import BaseCache
from chembl_beaker.beaker import config
from chembl_beaker.beaker.utils.functional import cached_property

#-----------------------------------------------------------------------------------------------------------------------

class BaseMemcachedCache(BaseCache):
    def __init__(self, library, value_not_found_exception):
        super(BaseMemcachedCache, self).__init__()
        if config.get('memcached_servers'):
            self._servers = json.loads(config.get('memcached_servers'))
        else:
            self._servers = [config.get('memcached_server', '127.0.0.1:11211')]

        # The exception type to catch from the underlying library for a key
        # that was not found. This is a ValueError for python-memcache,
        # pylibmc.NotFound for pylibmc, and cmemcache will return None without
        # raising an exception.
        self.LibraryValueNotFoundException = value_not_found_exception

        self._lib = library

#-----------------------------------------------------------------------------------------------------------------------

    @property
    def _cache(self):
        """
        Implements transparent thread-safe access to a memcached client.
        """
        if getattr(self, '_client', None) is None:
            self._client = self._lib.Client(self._servers)

        return self._client

#-----------------------------------------------------------------------------------------------------------------------

    def add(self, key, value):
        key = self.make_key(key)
        return self._cache.add(key, value, self.default_timeout)

#-----------------------------------------------------------------------------------------------------------------------

    def get(self, key, default=None):
        key = self.make_key(key)
        val = self._cache.get(key)
        if val is None:
            return default
        return val

#-----------------------------------------------------------------------------------------------------------------------

    def set(self, key, value):
        key = self.make_key(key)
        self._cache.set(key, value, self.default_timeout)

#-----------------------------------------------------------------------------------------------------------------------

    def delete(self, key):
        key = self.make_key(key)
        self._cache.delete(key)

#-----------------------------------------------------------------------------------------------------------------------

    def close(self, **kwargs):
        self._cache.disconnect_all()

#-----------------------------------------------------------------------------------------------------------------------

    def clear(self):
        self._cache.flush_all()


#-----------------------------------------------------------------------------------------------------------------------

class MemcachedCache(BaseMemcachedCache):
    "An implementation of a cache binding using python-memcached"
    def __init__(self):
        import memcache
        super(MemcachedCache, self).__init__(library=memcache, value_not_found_exception=ValueError)

    @property
    def _cache(self):
        if getattr(self, '_client', None) is None:
            self._client = self._lib.Client(self._servers, pickleProtocol=pickle.HIGHEST_PROTOCOL)
        return self._client


#-----------------------------------------------------------------------------------------------------------------------

class PyLibMCCache(BaseMemcachedCache):
    "An implementation of a cache binding using pylibmc"
    def __init__(self):
        import pylibmc
        super(PyLibMCCache, self).__init__(library=pylibmc, value_not_found_exception=pylibmc.NotFound)

    @cached_property
    def _cache(self):
        client = self._lib.Client(self._servers)
        return client

#-----------------------------------------------------------------------------------------------------------------------



__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

import json
from bottle import request, response
from netaddr import IPNetwork, IPAddress
from chembl_beaker.beaker import config

#-----------------------------------------------------------------------------------------------------------------------

IP_WHITELIST = config.get('ip_whitelist')

try:
    if IP_WHITELIST:
        IP_WHITELIST = [IPNetwork(ip) for ip in json.loads(IP_WHITELIST)]
except:
    pass

IP_BLACKLIST = config.get('ip_blacklist')

try:
    if IP_BLACKLIST:
        IP_BLACKLIST = [IPNetwork(ip) for ip in json.loads(IP_BLACKLIST)]
except:
    pass

#-----------------------------------------------------------------------------------------------------------------------

class Restrictions(object):
    name = 'restrictions'
    api = 2

    def apply(self, fn, context):
        def _check_restrictions(*args, **kwargs):
            if config.get('request_max_size') and request.content_length > int(config['request_max_size']):
                response.status = 413
                response.body = 'Request size larger than %s bytes' % config['request_max_size']
                return response

            ip = IPAddress(request.remote_addr)
            if IP_WHITELIST:
                if not any([ip in white for white in IP_WHITELIST]):
                    response.status = 403
                    return response
            elif IP_BLACKLIST:
                if any([ip in black for black in IP_BLACKLIST]):
                    response.status = 403
                    return response

            return fn(*args, **kwargs)

        return _check_restrictions

#-----------------------------------------------------------------------------------------------------------------------


# $Id$
#
#  Copyright (C) 2008-2011 Greg Landrum
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
from rdkit import Chem
import numpy
import math
import copy

periodicTable=Chem.GetPeriodicTable()

class Font(object):
  face='sans'
  size='12'
  weight='normal'
  name=None
  def __init__(self,face=None,size=None,name=None,weight=None):
    if face: self.face=face
    if size: self.size=size
    if name: self.name=name
    if weight: self.weight=weight

class DrawingOptions(object):
  dotsPerAngstrom= 30
  useFraction= 0.85
  
  atomLabelFontFace= "sans"
  atomLabelFontSize= 12
  atomLabelMinFontSize= 7

  bondLineWidth= 1.2
  dblBondOffset= .25
  dblBondLengthFrac= .8

  defaultColor= (1,0,0)
  selectColor= (1,0,0)
  bgColor = (1,1,1)

  colorBonds= True
  noCarbonSymbols= True
  includeAtomNumbers= False
  atomNumberOffset= 0
  radicalSymbol= u'\u2219'

  dash= (4,4)

  wedgeDashedBonds= True

  # used to adjust overall scaling for molecules that have been laid out with non-standard
  # bond lengths
  coordScale= 1.0

  elemDict={
    1:(0.55,0.55,0.55),
    7:(0,0,1),
    8:(1,0,0),
    9:(.2,.8,.8),
    15:(1,.5,0),
    16:(.8,.8,0),
    17:(0,.8,0),
    35:(.5,.3,.1),
    0:(.5,.5,.5),
    }

class MolDrawing(object):
  atomPs = None
  canvas = None
  canvasSize = None

  def __init__(self,canvas=None, drawingOptions=None):
    self.canvas = canvas
    if canvas:
      self.canvasSize=canvas.size
    self.atomPs = {}
    if drawingOptions is None:
      self.drawingOptions=DrawingOptions()
    else:
      self.drawingOptions=drawingOptions
    self.boundingBoxes = {}

    if self.drawingOptions.bgColor is not None:
      self.canvas.addCanvasPolygon(((0,0),
                                    (canvas.size[0],0),
                                    (canvas.size[0],canvas.size[1]),
                                    (0,canvas.size[1])),
                                   color=self.drawingOptions.bgColor,
                                   fill=True,stroke=False)

    
  def transformPoint(self,pos):
    res = [0,0]
    res[0] = (pos[0] + self.molTrans[0])*self.currDotsPerAngstrom*self.drawingOptions.useFraction + self.drawingTrans[0]
    res[1] = self.canvasSize[1]-((pos[1] + self.molTrans[1])*self.currDotsPerAngstrom*self.drawingOptions.useFraction + \
                                 self.drawingTrans[1])
    return res
  

  def _getBondOffset(self,p1,p2):
    # get the vector between the points:
    dx = p2[0]-p1[0]
    dy = p2[1]-p1[1]

    # figure out the angle and the perpendicular:
    ang = math.atan2(dy,dx)
    perp = ang + math.pi/2.

    # here's the offset for the parallel bond:
    offsetX = math.cos(perp)*self.drawingOptions.dblBondOffset*self.currDotsPerAngstrom
    offsetY = math.sin(perp)*self.drawingOptions.dblBondOffset*self.currDotsPerAngstrom

    return perp,offsetX,offsetY

  def _getOffsetBondPts(self,p1,p2,
                        offsetX,offsetY,
                        lenFrac=None):
    if not lenFrac:
      lenFrac = self.drawingOptions.dblBondLengthFrac
      
    dx = p2[0]-p1[0]
    dy = p2[1]-p1[1]
    # ----
    # now figure out where to start and end it:
    
    # offset the start point:
    fracP1 = p1[0]+offsetX,p1[1]+offsetY

    # now move a portion of the way along the line to the neighbor:
    frac = (1.-lenFrac)/2
    fracP1 = fracP1[0]+dx*frac,\
             fracP1[1]+dy*frac

    fracP2 = fracP1[0]+dx*lenFrac,\
             fracP1[1]+dy*lenFrac
    return fracP1,fracP2

  def _offsetDblBond(self,p1,p2,bond,a1,a2,conf,dir=1,
                     lenFrac=None):
    perp,offsetX,offsetY = self._getBondOffset(p1,p2)
    offsetX = offsetX*dir
    offsetY = offsetY*dir
    
    # if we're a ring bond, we may need to flip over to the other side:
    if bond.IsInRing():
      bondIdx = bond.GetIdx()
      a1Idx = a1.GetIdx()
      a2Idx = a2.GetIdx()
      # find a ring bond from a1 to an atom other than a2:
      for otherBond in a1.GetBonds():
        if otherBond.GetIdx()!=bondIdx and \
           otherBond.IsInRing():
          sharedRing=False
          for ring in self.bondRings:
            if bondIdx in ring and otherBond.GetIdx() in ring:
              sharedRing=True
              break
          if not sharedRing:
            continue
          a3 = otherBond.GetOtherAtom(a1)
          if a3.GetIdx() != a2Idx:
            p3 = self.transformPoint(conf.GetAtomPosition(a3.GetIdx())*self.drawingOptions.coordScale)
            dx2 = p3[0] - p1[0]
            dy2 = p3[1] - p1[1]
            dotP = dx2*offsetX + dy2*offsetY
            if dotP < 0:
              perp += math.pi
              offsetX = math.cos(perp)*self.drawingOptions.dblBondOffset*self.currDotsPerAngstrom
              offsetY = math.sin(perp)*self.drawingOptions.dblBondOffset*self.currDotsPerAngstrom
            
    fracP1,fracP2 = self._getOffsetBondPts(p1,p2,
                                           offsetX,offsetY,
                                           lenFrac=lenFrac)
    return fracP1,fracP2

  def _getBondAttachmentCoordinates(self, p1, p2, labelSize):
    newpos = [None, None]
    if labelSize != None:
      labelSizeOffset = [labelSize[0][0]/2 + (cmp(p2[0], p1[0]) * labelSize[0][2]), labelSize[0][1]/2]
      if p1[1] == p2[1]:
        newpos[0] = p1[0] + cmp(p2[0], p1[0]) * labelSizeOffset[0]
      else:
        if abs(labelSizeOffset[1] * (p2[0] - p1[0]) / (p2[1] - p1[1])) < labelSizeOffset[0]:
          newpos[0] = p1[0] + cmp(p2[0], p1[0]) * abs(labelSizeOffset[1] * (p2[0] - p1[0]) / (p2[1] - p1[1]))
        else:
          newpos[0] = p1[0] + cmp(p2[0], p1[0]) * labelSizeOffset[0]
      if p1[0] == p2[0]:
        newpos[1] = p1[1] + cmp(p2[1], p1[1]) * labelSizeOffset[1]
      else:
        if abs(labelSizeOffset[0] * (p1[1] - p2[1]) / (p2[0] - p1[0])) < labelSizeOffset[1]:
          newpos[1] = p1[1] + cmp(p2[1], p1[1]) * abs(labelSizeOffset[0] * (p1[1] - p2[1]) / (p2[0] - p1[0]))
        else:
          newpos[1] = p1[1] + cmp(p2[1], p1[1]) * labelSizeOffset[1]
    else:
      newpos = copy.deepcopy(p1)
    return newpos

  def _drawWedgedBond(self,bond,pos,nbrPos,
                      width=None,color=None,
                      dash=None):
    if width is None:
        width = self.drawingOptions.bondLineWidth
    if color is None:
        color = self.drawingOptions.defaultColor
    perp,offsetX,offsetY = self._getBondOffset(pos,nbrPos)
    offsetX *=.75
    offsetY *=.75
    poly = ((pos[0],pos[1]),
            (nbrPos[0]+offsetX,nbrPos[1]+offsetY),
            (nbrPos[0]-offsetX,nbrPos[1]-offsetY))
    #canvas.drawPolygon(poly,edgeColor=color,edgeWidth=1,fillColor=color,closed=1)
    if not dash:
      self.canvas.addCanvasPolygon(poly,color=color)
    elif self.drawingOptions.wedgeDashedBonds and self.canvas.addCanvasDashedWedge:
      self.canvas.addCanvasDashedWedge(poly[0],poly[1],poly[2],color=color)
    else:
      self.canvas.addCanvasLine(pos,nbrPos,linewidth=width*2,color=color,
                    dashes=dash)
    
  def _drawBond(self,bond,atom,nbr,pos,nbrPos,conf,
                width=None,color=None,color2=None,labelSize1=None,labelSize2=None):
    if width is None:
      width = self.drawingOptions.bondLineWidth
    if color is None:
      color = self.drawingOptions.defaultColor
    p1_raw = copy.deepcopy(pos)
    p2_raw = copy.deepcopy(nbrPos)
    newpos = self._getBondAttachmentCoordinates(p1_raw, p2_raw, labelSize1)
    newnbrPos = self._getBondAttachmentCoordinates(p2_raw, p1_raw, labelSize2)
    bType=bond.GetBondType()
    if bType == Chem.BondType.SINGLE:
      bDir = bond.GetBondDir()
      if bDir in (Chem.BondDir.BEGINWEDGE,Chem.BondDir.BEGINDASH):
        # if the bond is "backwards", change the drawing direction:
        if bond.GetBeginAtom().GetChiralTag() in (Chem.ChiralType.CHI_TETRAHEDRAL_CW,
                                                  Chem.ChiralType.CHI_TETRAHEDRAL_CCW):
          p1,p2 = newpos,newnbrPos
          wcolor=color
        else:
          p2,p1 = newpos,newnbrPos
          if color2 is not None:
            wcolor=color2
          else:
            wcolor=self.drawingOptions.defaultColor
        if bDir==Chem.BondDir.BEGINWEDGE:
          self._drawWedgedBond(bond,p1,p2,color=wcolor,width=width)
        elif bDir==Chem.BondDir.BEGINDASH:
          self._drawWedgedBond(bond,p1,p2,color=wcolor,width=width,
                               dash=self.drawingOptions.dash)

      else:
        self.canvas.addCanvasLine(newpos, newnbrPos, linewidth=width, color=color, color2=color2)
    elif bType == Chem.BondType.DOUBLE:
      if bond.IsInRing() or (atom.GetDegree()!=1 and bond.GetOtherAtom(atom).GetDegree()!=1):
        self.canvas.addCanvasLine(newpos,newnbrPos,linewidth=width,color=color,color2=color2)
        fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf)
        self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2)
      else:
        fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf,dir=.5,
                                      lenFrac=1.0)
        self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2)
        fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf,dir=-.5,
                                      lenFrac=1.0)
        self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2)
    elif bType == Chem.BondType.AROMATIC:
      self.canvas.addCanvasLine(newpos,newnbrPos,linewidth=width,color=color,color2=color2)
      fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf)
      self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2,
                    dash=self.drawingOptions.dash)
    elif bType == Chem.BondType.TRIPLE:
      self.canvas.addCanvasLine(newpos,newnbrPos,linewidth=width,color=color,color2=color2)
      fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf)
      self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2)
      fp1,fp2 = self._offsetDblBond(newpos,newnbrPos,bond,atom,nbr,conf,dir=-1)
      self.canvas.addCanvasLine(fp1,fp2,linewidth=width,color=color,color2=color2)
    else:
      self.canvas.addCanvasLine(newpos, newnbrPos, linewidth=width, color=color, color2=color2,
                                dash=(1,2))
      
  def scaleAndCenter(self,mol,conf,coordCenter=False,canvasSize=None,ignoreHs=False):
    if canvasSize is None:
      canvasSize=self.canvasSize
    xAccum = 0
    yAccum = 0
    minX = 1e8
    minY = 1e8
    maxX = -1e8
    maxY = -1e8

    nAts = mol.GetNumAtoms()
    for i in range(nAts):
      if ignoreHs and mol.GetAtomWithIdx(i).GetAtomicNum()==1: continue
      pos = conf.GetAtomPosition(i)*self.drawingOptions.coordScale
      xAccum += pos[0]
      yAccum += pos[1]
      minX = min(minX,pos[0])
      minY = min(minY,pos[1])
      maxX = max(maxX,pos[0])
      maxY = max(maxY,pos[1])

    dx = abs(maxX-minX)
    dy = abs(maxY-minY)
    xSize = dx*self.currDotsPerAngstrom
    ySize = dy*self.currDotsPerAngstrom

    if coordCenter:
      molTrans = -xAccum/nAts,-yAccum/nAts
    else:
      molTrans = -(minX+(maxX-minX)/2),-(minY+(maxY-minY)/2)
    self.molTrans = molTrans

    if xSize>=.95*canvasSize[0]:
      scale = .9*canvasSize[0]/xSize
      xSize*=scale
      ySize*=scale
      self.currDotsPerAngstrom*=scale
      self.currAtomLabelFontSize = max(self.currAtomLabelFontSize*scale,
                                       self.drawingOptions.atomLabelMinFontSize)
    if ySize>=.95*canvasSize[1]:
      scale = .9*canvasSize[1]/ySize
      xSize*=scale
      ySize*=scale
      self.currDotsPerAngstrom*=scale
      self.currAtomLabelFontSize = max(self.currAtomLabelFontSize*scale,
                                       self.drawingOptions.atomLabelMinFontSize)
    drawingTrans = canvasSize[0]/2,canvasSize[1]/2
    self.drawingTrans = drawingTrans

  def _drawLabel(self,label,pos,baseOffset,font,color=None,**kwargs):
    if color is None:
      color = self.drawingOptions.defaultColor
    x1 = pos[0]
    y1 = pos[1]
    labelP = x1,y1
    labelSize = self.canvas.addCanvasText(label,(x1,y1,baseOffset),font,color,**kwargs)
    return labelSize
    
  def AddMol(self,mol,centerIt=True,molTrans=None,drawingTrans=None,
             highlightAtoms=[],confId=-1,flagCloseContactsDist=2,
             highlightMap=None, ignoreHs=False,highlightBonds=[],**kwargs):
    """Set the molecule to be drawn.

    Parameters:
      hightlightAtoms -- list of atoms to highlight (default [])
      highlightMap -- dictionary of (atom, color) pairs (default None)

    Notes:
      - specifying centerIt will cause molTrans and drawingTrans to be ignored
    """
    conf = mol.GetConformer(confId)
    if 'coordScale' in kwargs:
      self.drawingOptions.coordScale=kwargs['coordScale']

    self.currDotsPerAngstrom=self.drawingOptions.dotsPerAngstrom
    self.currAtomLabelFontSize=self.drawingOptions.atomLabelFontSize
    if centerIt:
      self.scaleAndCenter(mol,conf,ignoreHs=ignoreHs)
    else:
      if molTrans is None:
        molTrans = (0,0)
      self.molTrans = molTrans
      if drawingTrans is None:
        drawingTrans = (0,0)
      self.drawingTrans = drawingTrans

    font = Font(face=self.drawingOptions.atomLabelFontFace,size=self.currAtomLabelFontSize)

    obds=None
    if not mol.HasProp('_drawingBondsWedged'):
      # this is going to modify the molecule, get ready to undo that
      obds=[x.GetBondDir() for x in mol.GetBonds()]
      Chem.WedgeMolBonds(mol,conf)
      
    includeAtomNumbers = kwargs.get('includeAtomNumbers',self.drawingOptions.includeAtomNumbers)
    self.atomPs[mol] = {}
    self.boundingBoxes[mol] = [0]*4
    self.activeMol = mol
    self.bondRings = mol.GetRingInfo().BondRings()
    labelSizes = {}
    for atom in mol.GetAtoms():
      labelSizes[atom.GetIdx()] = None
      if ignoreHs and atom.GetAtomicNum()==1:
        drawAtom=False
      else:
        drawAtom=True
      idx = atom.GetIdx()
      pos = self.atomPs[mol].get(idx,None)
      if pos is None:
        pos = self.transformPoint(conf.GetAtomPosition(idx)*self.drawingOptions.coordScale)
        self.atomPs[mol][idx] = pos
        if drawAtom:
          self.boundingBoxes[mol][0]=min(self.boundingBoxes[mol][0],pos[0])
          self.boundingBoxes[mol][1]=min(self.boundingBoxes[mol][1],pos[1])
          self.boundingBoxes[mol][2]=max(self.boundingBoxes[mol][2],pos[0])
          self.boundingBoxes[mol][3]=max(self.boundingBoxes[mol][3],pos[1])

      if not drawAtom: continue
      nbrSum = [0,0]
      for bond in atom.GetBonds():
        nbr = bond.GetOtherAtom(atom)
        if ignoreHs and nbr.GetAtomicNum()==1: continue
        nbrIdx = nbr.GetIdx()
        if nbrIdx > idx:
          nbrPos = self.atomPs[mol].get(nbrIdx,None)
          if nbrPos is None:
            nbrPos = self.transformPoint(conf.GetAtomPosition(nbrIdx)*self.drawingOptions.coordScale)
            self.atomPs[mol][nbrIdx] = nbrPos
            self.boundingBoxes[mol][0]=min(self.boundingBoxes[mol][0],nbrPos[0])
            self.boundingBoxes[mol][1]=min(self.boundingBoxes[mol][1],nbrPos[1])
            self.boundingBoxes[mol][2]=max(self.boundingBoxes[mol][2],nbrPos[0])
            self.boundingBoxes[mol][3]=max(self.boundingBoxes[mol][3],nbrPos[1])
            
        else:
          nbrPos = self.atomPs[mol][nbrIdx]
        nbrSum[0] += nbrPos[0]-pos[0]
        nbrSum[1] += nbrPos[1]-pos[1]
  
      iso = atom.GetIsotope()

      labelIt= not self.drawingOptions.noCarbonSymbols or \
               atom.GetAtomicNum()!=6 or \
               atom.GetFormalCharge()!=0 or \
               atom.GetNumRadicalElectrons() or \
               includeAtomNumbers or \
               iso or \
               atom.HasProp('molAtomMapNumber') or \
               atom.GetDegree()==0
      orient=''
      if labelIt:
        baseOffset = 0
        if includeAtomNumbers:
          symbol = str(atom.GetIdx())
          symbolLength = len(symbol)
        else:
          base = atom.GetSymbol()
          symbolLength = len(base)  
          nHs = atom.GetTotalNumHs()
          if nHs>0:
            if nHs>1:
              hs='H<sub>%d</sub>'%nHs
              symbolLength += 1 + len(str(nHs))
            else:
              hs ='H'
              symbolLength += 1
          else:
            hs = ''
          chg = atom.GetFormalCharge()
          if chg!=0:
            if chg==1:
              chg = '+'
            elif chg==-1:
              chg = '-'
            elif chg>1:
              chg = '+%d'%chg
            elif chg<-1:
              chg = '-%d'%chg
            symbolLength += len(chg) 
          else:
            chg = ''
          if chg:
            chg = '<sup>%s</sup>'%chg

          if atom.GetNumRadicalElectrons():
            rad = self.drawingOptions.radicalSymbol*atom.GetNumRadicalElectrons()
            rad = '<sup>%s</sup>'%rad
            symbolLength += atom.GetNumRadicalElectrons()
          else:
            rad = ''

          isotope=''
          isotopeLength = 0
          if iso:
            isotope='<sup>%d</sup>'%atom.GetIsotope()
            isotopeLength = len(str(atom.GetIsotope()))
            symbolLength += isotopeLength
          mapNum=''
          mapNumLength = 0
          if atom.HasProp('molAtomMapNumber'):
            mapNum=':'+atom.GetProp('molAtomMapNumber')
            mapNumLength = 1 + len(str(atom.GetProp('molAtomMapNumber')))
            symbolLength += mapNumLength
          deg = atom.GetDegree()
          # This should be done in a better way in the future:
          # 'baseOffset' should be determined by getting the size of 'isotope' and the size of 'base', or the size of 'mapNum' and the size of 'base'
          # (depending on 'deg' and 'nbrSum[0]') in order to determine the exact position of the base
          if deg==0:
            if periodicTable.GetElementSymbol(atom.GetAtomicNum()) in ('O','S','Se','Te','F','Cl','Br','I','At'):
              symbol = '%s%s%s%s%s%s'%(hs,isotope,base,chg,rad,mapNum)
            else:
              symbol = '%s%s%s%s%s%s'%(isotope,base,hs,chg,rad,mapNum)
          elif deg>1 or nbrSum[0]<1:
            symbol = '%s%s%s%s%s%s'%(isotope,base,hs,chg,rad,mapNum)
            baseOffset = 0.5 - (isotopeLength + len(base) / 2.) / symbolLength
          else:
            symbol = '%s%s%s%s%s%s'%(rad,chg,hs,isotope,base,mapNum)
            baseOffset = -0.5 + (mapNumLength + len(base) / 2.) / symbolLength
          if deg==1:
            if abs(nbrSum[1])>1:
              islope=nbrSum[0]/abs(nbrSum[1])
            else:
              islope=nbrSum[0]
            if abs(islope)>.3:
              if islope>0:
                orient='W'
              else:
                orient='E'
            elif abs(nbrSum[1])>10:
              if nbrSum[1]>0:
                orient='N'
              else :
                orient='S'
          else:
            orient = 'C'
        if highlightMap and idx in highlightMap:
          color = highlightMap[idx]
        elif highlightAtoms and idx in highlightAtoms:
          color = self.drawingOptions.selectColor
        else:
          color = self.drawingOptions.elemDict.get(atom.GetAtomicNum(),(0,0,0))
        labelSize = self._drawLabel(symbol, pos, baseOffset, font, color=color,orientation=orient)
        labelSizes[atom.GetIdx()] = [labelSize, orient]

    for bond in mol.GetBonds():
          atom, idx = bond.GetBeginAtom(), bond.GetBeginAtomIdx()
          nbr, nbrIdx = bond.GetEndAtom(), bond.GetEndAtomIdx()
          pos = self.atomPs[mol].get(idx,None)
          nbrPos = self.atomPs[mol].get(nbrIdx,None)
          if highlightBonds and bond.GetIdx() in highlightBonds:
            width=2.0*self.drawingOptions.bondLineWidth
            color = self.drawingOptions.selectColor
            color2 = self.drawingOptions.selectColor
          elif highlightAtoms and idx in highlightAtoms and nbrIdx in highlightAtoms:
            width=2.0*self.drawingOptions.bondLineWidth
            color = self.drawingOptions.selectColor
            color2 = self.drawingOptions.selectColor
          elif highlightMap is not None and idx in highlightMap and nbrIdx in highlightMap:
            width=2.0*self.drawingOptions.bondLineWidth
            color = highlightMap[idx]
            color2 = highlightMap[nbrIdx]
          else:
            width=self.drawingOptions.bondLineWidth
            if self.drawingOptions.colorBonds:
              color = self.drawingOptions.elemDict.get(atom.GetAtomicNum(),(0,0,0))
              color2 = self.drawingOptions.elemDict.get(nbr.GetAtomicNum(),(0,0,0))
            else:
              color = self.drawingOptions.defaultColor
              color2= color
          self._drawBond(bond,atom,nbr,pos,nbrPos,conf,
                           color=color,width=width,color2=color2,labelSize1=labelSizes[idx],labelSize2=labelSizes[nbrIdx])
  
    # if we modified the bond wedging state, undo those changes now
    if obds:
      for i,d in enumerate(obds):
        mol.GetBondWithIdx(i).SetBondDir(d)
      
      
    if flagCloseContactsDist>0:
      tol = flagCloseContactsDist*flagCloseContactsDist
      for i,atomi in enumerate(mol.GetAtoms()):
        pi = numpy.array(self.atomPs[mol][i])
        for j in range(i+1,mol.GetNumAtoms()):
          pj = numpy.array(self.atomPs[mol][j])
          d = pj-pi
          dist2 = d[0]*d[0]+d[1]*d[1]
          if dist2<=tol:
            self.canvas.addCanvasPolygon(((pi[0]-2*flagCloseContactsDist,
                                      pi[1]-2*flagCloseContactsDist),
                                     (pi[0]+2*flagCloseContactsDist,
                                      pi[1]-2*flagCloseContactsDist),
                                     (pi[0]+2*flagCloseContactsDist,
                                      pi[1]+2*flagCloseContactsDist),
                                     (pi[0]-2*flagCloseContactsDist,
                                      pi[1]+2*flagCloseContactsDist)),
                             color=(1.,0,0),
                             fill=False,stroke=True)



__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

import time
import hashlib
from chembl_beaker.beaker.utils import import_class
from chembl_beaker.beaker import config

#-----------------------------------------------------------------------------------------------------------------------

def default_key_func(key, key_prefix):
    """
    Default function to generate keys.

    Constructs the key used by all other methods. By default it prepends
    the `key_prefix'. KEY_FUNCTION can be used to specify an alternate
    function with custom key making behavior.
    """
    return hashlib.md5(':'.join([key_prefix, key])).hexdigest()

#-----------------------------------------------------------------------------------------------------------------------

def get_key_func(key_func):
    """
    Function to decide which key function to use.

    Defaults to ``default_key_func``.
    """
    if key_func is not None:
        if callable(key_func):
            return key_func
        else:
            return import_class(key_func)
    return default_key_func

#-----------------------------------------------------------------------------------------------------------------------

class BaseCache(object):

    def __init__(self):
        timeout = config.get('cache_timeout', 300)
        if timeout is not None:
            try:
                timeout = int(timeout)
            except (ValueError, TypeError):
                timeout = 300
            self.default_timeout = timeout
        max_entries = config.get('cache_max_entries', 300000)
        try:
            self._max_entries = int(max_entries)
        except (ValueError, TypeError):
            self._max_entries = 300
        cull_frequency = config.get('cache_cull_frequency', 3)
        try:
            self._cull_frequency = int(cull_frequency)
        except (ValueError, TypeError):
            self._cull_frequency = 3
        self.key_prefix = config.get('cache_key_prefix', '')
        self.key_func = get_key_func(config.get('cache_key_function', None))

#-----------------------------------------------------------------------------------------------------------------------

    def make_key(self, key):
        """Constructs the key used by all other methods. By default it
        uses the key_func to generate a key (which, by default,
        prepends the `key_prefix' and 'version'). An different key
        function can be provided at the time of cache construction;
        alternatively, you can subclass the cache backend to provide
        custom key making behavior.
        """

        new_key = self.key_func(key, self.key_prefix)
        return new_key

#-----------------------------------------------------------------------------------------------------------------------

    def add(self, key, value):
        """
        Set a value in the cache if the key does not already exist. If
        timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        Returns True if the value was stored, False otherwise.
        """
        raise NotImplementedError('subclasses of BaseCache must provide an add() method')

#-----------------------------------------------------------------------------------------------------------------------

    def get(self, key, default=None):
        """
        Fetch a given key from the cache. If the key does not exist, return
        default, which itself defaults to None.
        """
        raise NotImplementedError('subclasses of BaseCache must provide a get() method')

#-----------------------------------------------------------------------------------------------------------------------

    def set(self, key, value):
        """
        Set a value in the cache. If timeout is given, that timeout will be
        used for the key; otherwise the default cache timeout will be used.
        """
        raise NotImplementedError('subclasses of BaseCache must provide a set() method')

#-----------------------------------------------------------------------------------------------------------------------

    def delete(self, key):
        """
        Delete a key from the cache, failing silently.
        """
        raise NotImplementedError('subclasses of BaseCache must provide a delete() method')

#-----------------------------------------------------------------------------------------------------------------------

    def clear(self):
        """Remove *all* values from the cache at once."""
        raise NotImplementedError('subclasses of BaseCache must provide a clear() method')

#-----------------------------------------------------------------------------------------------------------------------

    def get_backend_timeout(self):
        """
        Returns the timeout value usable by this backend based upon the provided
        timeout.
        """
        return time.time() + self.default_timeout

#-----------------------------------------------------------------------------------------------------------------------